//##############################################################################################
// Title	:	NOKIA_3310.c								
// Version	:	v1.4									
// Date		:	Januar 2010						
// Info		:	Nokia 3310 Lcd Ansteuerung
// Author	:	Martin Junghans, Vorlagen von Hagen Redmann(Kreisfunktionen), 
//				und einigen anderen Codeschnipseln
// Compiler	:	AVR-GCC - WinAVR
//##############################################################################################
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include "Nokia_3310.h"
#include <util/delay.h>
#include <compat/deprecated.h>

#define sbi(ADDRESS,BIT) ((ADDRESS) |= (1<<(BIT)))
#define cbi(ADDRESS,BIT) ((ADDRESS) &= ~(1<<(BIT)))

/*SPI2X SPR1 SPR0 SCK Frequenz
	0     0    0    fosz/4
	0     0    1    fosz/16
	0     1    0    fosz/64
	0     1    1    fosz/128
	1     0    0    fosz/2
	1     0    1    fosz/8
	1     1    0    fosz/32
	1     1    1    fosz/64
*/
// GLOBALE VARIABLEN, benutzervernderbar
uint8_t 	*FontPointer; 	//FontPointer auf Font
uint8_t 	rot				=0;	//Ausrichtung der Schrift 0=0, 1=90
uint8_t 	TextFormat		=0;	//links=0 zentriert=1 rechts=2
uint8_t 	TextFarbe;		//Textfarbe, 5-6-5 RGB
uint8_t 	TextTyp;		//Texttyp 0=Proportional , 1=Fixed	
uint8_t		FNull=0;			// Fhrende Nullen bei Zahlen anzeigen	0=Aus , 1=an

// GLOBAELE VARIABLEN, zur internen Benutzung
uint8_t		TextBreite;		// Maximale Textbreite der aktuellen Font
//uint8_t		TextHhe;		// Maximale Textbreite der aktuellen Font
char		za[11];			// Zahlenzwischenspeicher
// GLOBALE VARIABLEN ende



static uint8_t N3310_Cache[N3310_CACHE_SIZE];

// HG-Bel-Wertetabelle
prog_uint8_t HG[11] = {0,60,100,130,150,170,180,185,190,195,200};
//##############################################################################################
void N3310_SetBackLightInit(uint8_t Helligkeit)
{
	cbi(PORTB,2); 	// clear port before enable
	sbi(DDRB, 2);	// will be used for OC2, must be output

	
	TCCR1A  |=  (1<<COM1B1)| (1<<WGM11);							// pin B5 und PWM mode	--> timer 1; WGM11=9Bit mode; com1b1=nicht invertiert; 
	TCCR1B 	|=   _BV(WGM13)| _BV(WGM12) |_BV(CS10)|_BV(CS11);		// timer1 prescaller 1023; _BV(CS10)|  _BV(CS12)|
	ICR1 	=	1000;
	OCR1B 	= 	ICR1/2;
	OCR1B = Helligkeit*10;

}

//##############################################################################################
void N3310_SetBackLight(uint8_t Helligkeit)	{OCR1B = Helligkeit*10;}

//#############################################################################
void N3310_spiInit(void)
{
	// Init Serial Ports
	sbi(N3310_SER_PORT, N3310_SCLK);
	sbi(N3310_SER_DDR, 	N3310_SCLK);
	sbi(N3310_SER_DDR, 	N3310_MOSI);
	sbi(N3310_SER_DDR, 	N3310_CS_);			// hardware
	// Init Hardware SPI
	SPCR = (1<<MSTR)|(1<<SPE);//|(1<<SPR0);
	//SPSR |= (1<<SPI2X);			//double speed bit
}
//#############################################################################
void N3310_spiSendByte(uint8_t data)
{
	cbi(N3310_CTRL_PORT, N3310_CS);			// Chip Enable
	SPDR = data;								// Send Byte
	while(!(SPSR & (1<<SPIF)));				// Wait until transfer complete
	sbi(N3310_CTRL_PORT, N3310_CS);			// Chip disable
}

//#############################################################################
void N3310_wrdata(uint8_t myData)
{
	sbi(N3310_CTRL_PORT, N3310_DC);				// Enable Data Transfer
	N3310_spiSendByte(myData);					// Send Byte
}

//#############################################################################
void N3310_wrcmd(uint8_t myData)
{
	cbi(N3310_CTRL_PORT, N3310_DC);				// Enable Control Transfer
	N3310_spiSendByte(myData);					// Send Byte
}

//#############################################################################
void N3310_InitDisplay(void)
{
	sbi(N3310_CTRL_DDR,  	N3310_DC);				// Init Control Ports

	sbi(N3310_CTRL_DDR,	 	N3310_CS);
	sbi(N3310_CTRL_PORT, 	N3310_CS);				// deselect Control Transfer

	sbi(N3310_CTRL_DDR, 	N3310_RES_);
	cbi(N3310_CTRL_PORT, 	N3310_RES_);			// Reset Lcd
	_delay_ms(1);
	sbi(N3310_CTRL_PORT, 	N3310_RES_);			// _delay_ms(100);

	
	N3310_wrcmd(N3310_FUNCTIONSETEXT);			// N3310_FUNCTIONSETEXT	0x21	Enable extended Functionset
	N3310_wrcmd(N3310_TEMPCOEF);				// N3310_TEMPCOEF		0x06	Set Tempcoefficient
	N3310_wrcmd(N3310_BIAS);					// N3310_BIAS			0x13	Set Bias
	N3310_Contrast(0x40);						// Set Contrast
	N3310_wrcmd(N3310_FUNCTIONSET);				// N3310_FUNCTIONSET	0x20	Enable normal Functionset
	N3310_wrcmd(N3310_DISPLAYNORMAL);			// N3310_DISPLAYNORMAL	0x0C	Set Display to normal Mode
	N3310_ClearLcd();							// Clear Lcd
}
//#############################################################################
void N3310_Contrast ( uint8_t contrast )
{
	N3310_wrcmd(N3310_FUNCTIONSETEXT);
	N3310_wrcmd(N3310_SETVOP | contrast);
	N3310_wrcmd(N3310_FUNCTIONSET);
}
//#############################################################################
void N3310_SetCharAddrXY(uint8_t myX, uint8_t myY)
{
	N3310_wrcmd(myX | N3310_SETXADDR);
	N3310_wrcmd(myY | N3310_SETYADDR);
}


//#############################################################################
void N3310_SetInvLcd(uint8_t myInv)
{
	if(myInv == 1)
		N3310_wrcmd(N3310_DISPLAYINVERT);
	if(myInv == 0)
		N3310_wrcmd(N3310_DISPLAYNORMAL);
}

//##############################################################################################
void N3310_DrawLine(uint8_t x1,uint8_t y1,uint8_t x2,uint8_t y2,uint8_t N3310_FgColor)
{
		if ((x1 == x2) | (y1 == y2)) 
			{
			N3310_DrawFilledRect(x1, y1, x2, y2, N3310_FgColor);
			} 
		else
			{
			N3310_Coord_t dx, dy, sx, sy;
			if (x1 < x2) 	{dx = x2 - x1; sx = +1;}
			else 			{dx = x1 - x2; sx = -1;}

			if (y1 < y2) 	{dy = y2 - y1; sy = +1;}
			else 			{dy = y1 - y2; sy = -1;}

			uint8_t swap = (dx > dy) ? 0 : 1;
			if (swap) 
				{
				N3310_SwapCoord(x1, y1);
				N3310_SwapCoord(dx, dy);
				N3310_SwapCoord(sx, sy);
				}

			int16_t e = dy - dx + dy;
			N3310_Coord_t x3 = x1;
			N3310_Coord_t y3 = y1;
			x2 = x1;
			y2 = y1;


			for (N3310_Coord_t i = 0; i <= dx; i++) 
				{
				if ((x1 != x2) & (y1 != y2)) 
					{
					if (swap) 
						{
						N3310_SwapCoord(x2, y2);
						N3310_SwapCoord(x3, y3);
						}
					N3310_DrawFilledRect(x2, y2, x3, y3, N3310_FgColor);
					x2 = x1;
					y2 = y1;
					};
				x3 = x1;
				y3 = y1;
				while (e >= 0) 
					{
					y1 += sy;
					e  -= dx;
					e  -= dx;
					}
				x1 += sx;
				e  += dy;
				e  += dy;
				}
			if (swap) 
				{
				N3310_SwapCoord(x2, y2);
				N3310_SwapCoord(x3, y3);
				}
			N3310_DrawFilledRect(x2, y2, x3, y3, N3310_FgColor);
			}
}
//############################################################################################## - fertig
void N3310_DrawFilledRect(uint8_t x1,uint8_t y1,uint8_t x2,uint8_t y2,uint8_t color)
{
	N3310_CheckCoord(x1, x2)
	N3310_CheckCoord(y1, y2)

	for (uint8_t y = y1; y <= y2; y++) 
		{
		for (uint8_t x = x1; x <= x2; x++)
			{
			N3310_SetPixel( x, y, color);
			}
		}
}
//##############################################################################################
void N3310_DrawRect(uint8_t x1,uint8_t y1,uint8_t x2,uint8_t y2,uint8_t color)
{
	N3310_CheckCoord(x1, x2)
	for(uint8_t i = x1; i <= x2; i++){      			// X-Seiten abf8x8hren
	N3310_SetPixel(i, y1, color);                        // obere H-Linie
	N3310_SetPixel(i, y2, color);                        // untere H-Linie
	}

	for(uint8_t i = y1; i <= y2; i++){               // Y-Seiten abf8x8hren
	N3310_SetPixel(x1, i, color);                        // linke V-Linie
	N3310_SetPixel(x2, i, color);                        // rechte V-Linie
	}
	
}

//##############################################################################################
void N3310_SetPixel4(N3310_Coord_t x1, N3310_Coord_t y1, N3310_Coord_t x2, N3310_Coord_t y2, uint8_t color) 
{
    N3310_SetPixel(x1, y1, color);
    N3310_SetPixel(x2, y1, color);
    N3310_SetPixel(x1, y2, color);
    N3310_SetPixel(x2, y2, color);
}

//##############################################################################################
void N3310_Line2(N3310_Coord_t x1, N3310_Coord_t y1, N3310_Coord_t x2, N3310_Coord_t y2, uint8_t color) 
{
    N3310_DrawFilledRect(x1, y1, x2, y1, color);
    N3310_DrawFilledRect(x1, y2, x2, y2, color);
}

//##############################################################################################
void N3310_RoundRect(N3310_Coord_t x1, N3310_Coord_t y1, N3310_Coord_t x2, N3310_Coord_t y2, N3310_Coord_t a, N3310_Coord_t b, uint8_t fgcolor, uint8_t bkcolor) 
{

	
	if (x1 > x2) N3310_SwapCoord(x1, x2);
	if (y1 > y2) N3310_SwapCoord(y1, y2);
    N3310_Coord_t t;
    t = x2 - x1;
    t >>= 1;
    if (a > t) a = t;
    t = y2 - y1;
    t >>= 1;
    if (b > t) b = t;
    x1 += a;
    x2 -= a;
    if (fgcolor != NONE) N3310_Line2(x1, y1, x2, y2, fgcolor);
    uint16_t as = a * a, bs = b * b;
    int32_t dx = 0, dy = as, ee = bs;
    dy *= b;
    ee += as / 4;
    ee -= dy;
    dy += dy;
	
    while (dx < dy) 
		{
		if (fgcolor != NONE) N3310_SetPixel4(x1, y1, x2, y2, fgcolor);
		x1--;
		x2++;
		if (ee >= 0) 
			{
			y1++;
			y2--;
			dy -= as;
			dy -= as;
			ee -= dy;
			if (bkcolor != NONE) N3310_Line2(x1, y1, x2, y2, bkcolor);
			}
		dx += bs;
		dx += bs;
		ee += bs;
		ee += dx;
		}

	int32_t 	tt = as;
				tt -= bs;
	tt *= 3;
	tt >>= 1;
	tt -= dx;
	tt -= dy;
	tt >>= 1;
    ee += tt;
	
    while (y1 <= y2) 
		{
		if (bkcolor != NONE) N3310_Line2(x1, y1, x2, y2, bkcolor);
		if (fgcolor != NONE) N3310_SetPixel4(x1, y1, x2, y2, fgcolor);
		y1++;
		y2--;
		if (ee < 0) 
			{
			x1--;
			x2++;
			a++;
			dx += bs;
			dx += bs;
			ee += dx;
			}
			dy -= as;
			dy -= as;
			ee += as;
			ee -= dy;
			}
}

//##############################################################################################
void N3310_Ellipse(N3310_Coord_t x1, N3310_Coord_t y1, N3310_Coord_t x2, N3310_Coord_t y2, uint16_t fgcolor, uint8_t bkcolor) 
{
    if (x1 > x2) N3310_SwapCoord(x1, x2);
    if (y1 > y2) N3310_SwapCoord(y1, y2);
    N3310_RoundRect(x1, y1, x2, y2, (x2 - x1) / 2, (y2 - y1) / 2, fgcolor, bkcolor);
}

//##############################################################################################
void N3310_Circle(N3310_Coord_t x, N3310_Coord_t y, N3310_Coord_t r, uint16_t fgcolor, uint16_t bkcolor) 
{
    N3310_RoundRect(x - r, y - r, x + r, y + r, r, r, fgcolor, bkcolor);
}



//##############################################################################################
// Routinen zur Modifikation der Umgebungsveriablen
void 	N3310_SetFont(uint8_t *pointer)			{FontPointer = pointer;TextBreite = pgm_read_byte_near(&FontPointer[2]);}
void 	N3310_SetRot(uint8_t r)					{rot = r;};
void 	N3310_SetTextFormat(uint8_t Format)		{TextFormat = Format;};
void 	N3310_SetTextColor(uint8_t color)			{TextFarbe = color;};
void 	N3310_SetTextTyp(uint8_t typ)				{TextTyp = typ;}
void 	N3310_SetFNull(uint8_t fn)				{FNull = fn;};


//#############################################################################################
uint8_t N3310_PutChar(uint8_t c) // Ein Zeichen ausgeben, vorher muss eine Schriftart geladen werden
{
	uint8_t 	hoehe 			=	pgm_read_byte_near(&FontPointer[3]);
	uint8_t 	firstchar 		=	pgm_read_byte_near(&FontPointer[5]);
	uint8_t 	lastchar 		=	pgm_read_byte_near(&FontPointer[6]);
	uint8_t 	zeichenbreite 	= 	pgm_read_byte_near(&FontPointer[7+c-firstchar]);
	uint8_t 	byte 			= 	0;
	uint8_t 	bitoffset 		= 	0;
	uint8_t 	maske 			= 	0;
	uint8_t 	bi 				= 	0;
	uint16_t 	bitsbischar 	= 	0;
	uint16_t 	bytesbischar	= 	0;
	uint16_t 	i 				= 	0;
	
	N3310_Coord_t x = N3310_Cursor.X;
    N3310_Coord_t y = N3310_Cursor.Y;

	for(uint8_t Ci = 0; Ci < c-firstchar; Ci++)
		{
		bitsbischar += (pgm_read_byte_near(&FontPointer[Ci+7]));
		}

	bitsbischar *= hoehe;
	bitsbischar += ((lastchar-firstchar)+7)*8;
	bytesbischar = bitsbischar/8;
	bitoffset = bitsbischar % 8;
	maske = bitoffset % 8;

	for(uint8_t xc = 0; xc < zeichenbreite; xc++) // Eigentliche Ausgaberoutine
		{
		for(uint8_t yc = hoehe; yc > 0; yc--)
			{
			if(maske > 7){maske = 0;bi+=1;};
			byte = pgm_read_byte_near(&FontPointer[bytesbischar + bi]+1);
			
			if(bit_is_set(byte,maske))
				{
				if(rot)
					{
						

					N3310_SetPixel(yc+y,xc+x, TextFarbe);
					}
				else
					{
					N3310_SetPixel(xc+x,(hoehe-yc)+y, TextFarbe);
					}
				}
			maske++;
			i++;
			}
		}
	return(zeichenbreite);
}

//##############################################################################################
uint16_t N3310_CharsWidth_Zahl(const char *addr)
{
	uint8_t 	firstchar 		=	pgm_read_byte_near(&FontPointer[5]);
	uint8_t 	c =0;;
    uint16_t 	i = 0;

    while (c = *addr++)
		{
		i += pgm_read_byte_near(&FontPointer[8+c-firstchar]); // Zeichenbreite addieren
		}
    return(i);
}
//##############################################################################################
uint16_t N3310_CharsWidth(const char *addr)
{
	uint8_t 	firstchar 		=	pgm_read_byte_near(&FontPointer[5]);
	uint8_t 	c =0;;
    uint16_t 	i = 0;

    while (c = pgm_read_byte(addr++))
		{
		i += pgm_read_byte_near(&FontPointer[8+c-firstchar]); // Zeichenbreite addieren
		}
    return(i);
}

//##############################################################################################
uint8_t	N3310_PutStr(char *c)
{	
	// NEW LINE ???
	//if (c == '\n') {      N3310_NewLine();      return(0);    }				// noch bearbeiten
	
	
	if(rot){                                          // x2 muss grer sein als x1
			N3310_Coord_t eX = N3310_Cursor.X;
			N3310_Cursor.X=N3310_Cursor.Y; N3310_Cursor.Y= eX;
			}


	N3310_Coord_t x = N3310_Cursor.X;
    N3310_Coord_t y = N3310_Cursor.Y;

	//TeXtausrichtung bzw format - anfang
	
	if (TextFormat == 1)	
		{		
		x -= N3310_CharsWidth(c)/ 2;
		y -= pgm_read_byte_near(&FontPointer[3])/ 2;	// Hhen Zentriert
		}	
	//if (TextFormat == 2)	{	x -= N3310_CharsWidth(c);	}
	//TeXtausrichtung bzw format  - ende

	uint8_t zeichen = 0;
	uint8_t zeichenbreite = 0;
	uint8_t xo = 0;
	uint8_t i = 0;
	while(zeichen = pgm_read_byte(c++))
		{
		N3310_Cursor.X=x+xo;
		zeichenbreite = N3310_PutChar(zeichen);
		if(TextTyp == 0)
			{
			xo += zeichenbreite+1;
			}
		else
			{
			xo+=TextBreite;
			}
		i++;
		if(i == 255){return(255);} // 255 ist "Fehleroode" fr unterminierten String
		}
	return(xo);
	}
//#############################################################################
int8_t	N3310_PutStr_zahl(char *c)
{	
	N3310_Coord_t x = N3310_Cursor.X;
    N3310_Coord_t y = N3310_Cursor.Y;
	//TeXtausrichtung bzw format - anfang
	if(rot)	{	N3310_SwapCoord(x, y) ;	}
	
	if (TextFormat == 1)	
		{
		x -= N3310_CharsWidth_Zahl(c)/ 2;
		//y -= pgm_read_byte_near(&FontPointer[3])/ 2;	// Hhen Zentriert
		}	
	if (TextFormat == 2)	
		{
		x -= N3310_CharsWidth_Zahl(c);
		}
	//TeXtausrichtung bzw format  - ende

	uint8_t zeichen = 0;
	uint8_t zeichenbreite = 0;
	uint8_t xo = 0;
	uint8_t i = 0;
	while(zeichen = *c++)//pgm_read_byte(c++))
		{
		N3310_Cursor.X=x+xo;
		zeichenbreite = N3310_PutChar(zeichen);
		if(TextTyp == 0)
			{
			xo += zeichenbreite+2;
			}
		else
			{
			xo+=TextBreite;
			}
		i++;
		if(i == 255){return(255);} // 255 ist "Fehleroode" fr unterminierten String
		}
	return(xo);
	}

//##############################################################################################
void N3310_PutInt(uint16_t z)// Routine zum ausgeben von Zahlen
{ 
	za[6] = 0;
	za[5] = (z % 10)+48;z = z/10;
	za[4] = (z % 10)+48;z = z/10;
	za[3] = (z % 10)+48;z = z/10;
	za[2] = (z % 10)+48;z = z/10;
	za[1] = (z % 10)+48;z = z/10;
	za[0] = (z)+48;
	
	uint8_t i = 0;
	if(!FNull)		{	while(za[i] == 48){i++;}		}	// Fhrende Nullen entfernen / lassen
	
	N3310_PutStr_zahl(&za[i]);
}

//##############################################################################################
void N3310_PutLong(uint32_t z) // Routine zum ausgeben von Zahlen
{
	za[10] = 0;
	for(uint8_t i = 9; i >0; i--)
		{
		za[i] =(z % 10)+48;z = z/10;
		}
	
	
	uint8_t i = 0;
	if(!FNull)										// Fhrende Nullen entfernen / lassen
		{
		while(za[i] == 48){i++;}
		}
	N3310_PutStr_zahl(&za[i]);
}
//############################################################################# Bild muss eine ganze page(8pixel) fllen (1-6)
void N3310_PutRawBmp16(uint8_t x0, uint8_t page0, uint8_t *pic, uint8_t breite, uint8_t hoehe)
{		
	uint16_t myIndex = 0;
	uint8_t page = hoehe/8;
	for(uint8_t myPage = page0; myPage < page0+page ; myPage++)
		{
		for (uint8_t x = x0; x < x0+breite ; x++)
			{
			N3310_Cache[(N3310_PIXELX*myPage)+x]=  pgm_read_byte(&pic[myIndex++]);
			}
			
		N3310_WriteCache(x0,myPage,breite);	
		}
}
/*
TO DO
momentan kann man Bilder in der Hhe nur in zeilen und nicht pixeln verschieben.
das muss noch gendert werden!
*/



/*	uint16_t myIndex = 0;
	uint8_t col;
	for(uint8_t page = 0; page < 6; page++)
	{
		for (col = 0; col < GLCD_PIXELX; col++)
			glcdCache[myIndex++] = PRG_RDB(&image[(GLCD_PIXELX*page)+col]);	// Cache
	}
	glcdn_WriteCache(0,0,504);
	*/
//#############################################################################
void N3310_SetPixel(uint8_t x, uint8_t y, uint8_t color)				// mode 0 = clear, mode 1 = draw
{

  if ((x >= N3310_Clip.X1) & (x <= N3310_Clip.X2) & (y >= N3310_Clip.Y1) & (y <= N3310_Clip.Y2)) 
	{
	uint8_t myPage = (y / 8);
	if(color == 0)		N3310_Cache[(N3310_PIXELX*myPage)+x] &= ~(1<<(y-(myPage*8)));
	if(color == 1)		N3310_Cache[(N3310_PIXELX*myPage)+x] |= (1<<(y-(myPage*8)));
	N3310_WriteCache(x,myPage,1);
	}
	
}

//#############################################################################
void N3310_WriteCache(uint8_t x, uint8_t page, uint16_t size)
{
	N3310_wrcmd(N3310_SETXADDR | x);
	N3310_wrcmd(N3310_SETYADDR | page);
	uint16_t myIndex = (page*N3310_PIXELX) + x;
	for(uint16_t i = 0;i<size;i++)
	{
		N3310_wrdata(N3310_Cache[myIndex + i]);
	}
}

//#############################################################################
void N3310_ClearLcd(void)
{
	for(uint8_t page = 0; page <= 6; page++)
		{
		for (uint8_t i = 0; i < N3310_PIXELX; i++)
			{
			N3310_wrdata( 0x00);
			}
		N3310_wrcmd(N3310_SETXADDR | 0);
		N3310_wrcmd(N3310_SETYADDR | page);
		}
	
	
	for (uint16_t i = 0; i < 504; i++)
		{
		N3310_Cache[i]=0x00;				
		}
}
//#############################################################################
